"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const utils_1 = require("./utils");
/**
 *
 * @param node
 * @param symbol
 * @param sourceFile
 * @param priorDiagnostics
 * @param log
 * @param declarationOverride  - allows us to override the declaration of the symbol if the deprecation was found elsewhere
 * @returns
 */
const createDeprecatedDiagnostic = (node, symbol, sourceFile, priorDiagnostics, log, declarationOverride) => {
    log(`Creating diagnostic for deprecated symbol: ${symbol.getName()} - ${node.getText()}`);
    const diagnosticStart = node.getStart();
    const diagnosticEnd = node.getEnd() - node.getStart();
    const diagnosticCode = 6385; // Deprecated code
    const declaration = declarationOverride || symbol.getDeclarations()[0];
    const deprecationSourceStart = declaration.getStart();
    const deprecationSourceEnd = declaration.getEnd();
    const deprecationSourceFile = declaration.getSourceFile();
    const alreadyReported = priorDiagnostics.some(diagnostic => diagnostic.start === diagnosticStart &&
        diagnostic.length === diagnosticEnd &&
        diagnostic.code === diagnosticCode);
    if (!alreadyReported) {
        log(`Adding diagnostic for deprecated symbol: ${symbol.getName()}`);
        return {
            file: sourceFile,
            start: diagnosticStart,
            length: diagnosticEnd,
            messageText: `'${symbol.getName()}' is deprecated.`,
            category: ts.DiagnosticCategory.Warning,
            reportsDeprecated: true,
            code: diagnosticCode,
            relatedInformation: [
                {
                    start: deprecationSourceStart,
                    length: deprecationSourceEnd - deprecationSourceStart,
                    file: deprecationSourceFile,
                    messageText: "The declaration was marked as deprecated here.",
                    category: ts.DiagnosticCategory.Error,
                    code: 2798
                }
            ]
        };
    }
    return null;
};
// Helper function to create detailed deprecation message for QuickInfo
const createDeprecatedQuickInfoTag = (symbol, checker) => {
    const declaration = symbol.getDeclarations()[0];
    const sourceFile = declaration.getSourceFile();
    const start = declaration.getStart();
    const end = declaration.getEnd();
    return {
        name: "deprecated",
        text: [
            {
                kind: "text",
                text: `Symbol '${symbol.getName()}' is deprecated.`
            },
            {
                kind: "space",
                text: " "
            },
            {
                kind: "text",
                text: `Declared in ${sourceFile.fileName} at line ${sourceFile.getLineAndCharacterOfPosition(start).line + 1}.`
            }
        ]
    };
};
// Traverse imports and re-exports to resolve and check for deprecations
const isImportDeclarationDeprecated = (currentNode, checker, log, program, optionalSymbolToMatch) => {
    var _a, _b;
    const diagnostics = [];
    // Get the import node for the current node
    let node = currentNode;
    // Handle Import Declarations
    if (ts.isImportDeclaration(node)) {
        log(`[IMPORT] Checking import declaration: ${node.getText()}`);
        const importClause = node.importClause;
        if (importClause && importClause.namedBindings) {
            const namedBindings = importClause.namedBindings;
            if (ts.isNamedImports(namedBindings)) {
                const moduleSpecifier = node.moduleSpecifier;
                const moduleName = moduleSpecifier.text;
                const sourceFile = node.getSourceFile();
                log(`[IMPORT] Resolving module specifier: "${moduleName}" from "${sourceFile.fileName}"`);
                // Resolve the module to get the file name
                const resolvedModule = ts.resolveModuleName(moduleName, sourceFile.fileName, program.getCompilerOptions(), ts.sys);
                if (resolvedModule.resolvedModule) {
                    const resolvedFileName = resolvedModule.resolvedModule.resolvedFileName;
                    log(`[IMPORT] Resolved module "${moduleName}" to "${resolvedFileName}"`);
                    // Skip node_modules
                    if (resolvedFileName.includes("node_modules")) {
                        log(`[IMPORT] Skipping node_modules`);
                        return [];
                    }
                    const moduleSourceFile = program.getSourceFile(resolvedFileName);
                    if (moduleSourceFile) {
                        // Get the module symbol from the source file
                        const moduleSymbol = checker.getSymbolAtLocation(moduleSourceFile);
                        if (moduleSymbol) {
                            const exports = checker.getExportsOfModule(moduleSymbol);
                            log(`[IMPORT] Found ${exports.length} exports in module "${moduleSourceFile.getSourceFile().fileName}"`);
                            for (const element of namedBindings.elements) {
                                const importName = element.name.text;
                                const exportSymbol = exports.find(s => s.name === importName);
                                // If the export symbol name doesn't match the current symbol, skip
                                if (optionalSymbolToMatch &&
                                    (exportSymbol === null || exportSymbol === void 0 ? void 0 : exportSymbol.getName()) !== optionalSymbolToMatch.getName()) {
                                    log(`[IMPORT] Skipping export "${importName} - ${exportSymbol === null || exportSymbol === void 0 ? void 0 : exportSymbol.getName()} - ${optionalSymbolToMatch.getName()}"`);
                                    continue;
                                }
                                const exportSymbolDeclaration = (_a = exportSymbol === null || exportSymbol === void 0 ? void 0 : exportSymbol.getDeclarations()) === null || _a === void 0 ? void 0 : _a[0];
                                /**
                                 * Check if the parent node of the export symbol is deprecated.
                                 */
                                let exportSymbolParent = exportSymbolDeclaration === null || exportSymbolDeclaration === void 0 ? void 0 : exportSymbolDeclaration.parent;
                                log(`[IMPORT] Checking parent node for deprecation: ${exportSymbolParent === null || exportSymbolParent === void 0 ? void 0 : exportSymbolParent.getText()}`);
                                while (exportSymbolParent && !ts.isSourceFile(exportSymbolParent)) {
                                    const isExportSymbolParentDeprecated = (0, utils_1.isNodeDeprecated)(checker, exportSymbolParent);
                                    log(`[IMPORT] Parent node is deprecated: ${isExportSymbolParentDeprecated}`);
                                    if (isExportSymbolParentDeprecated) {
                                        let importNode = undefined;
                                        node.forEachChild(child => {
                                            log(`[IMPORT] Child: ${child.getText()}`);
                                            child.forEachChild(namedImport => {
                                                log(`[IMPORT] Named Import: ${namedImport.getText()}`);
                                                namedImport.forEachChild(importSpec => {
                                                    log(`[IMPORT] Import Specifier: ${importSpec.getText()}`);
                                                    if (importSpec.getText() === importName) {
                                                        importNode = importSpec;
                                                    }
                                                });
                                            });
                                        });
                                        diagnostics.push(createDeprecatedDiagnostic(importNode, exportSymbol, exportSymbolDeclaration.getSourceFile(), [], log, exportSymbolDeclaration));
                                    }
                                    exportSymbolParent = exportSymbolParent === null || exportSymbolParent === void 0 ? void 0 : exportSymbolParent.parent;
                                }
                                if (exportSymbol) {
                                    log(`[IMPORT] Checking export "${importName}" in module "${moduleSourceFile.getSourceFile().fileName}"`);
                                    const declarations = exportSymbol.getDeclarations();
                                    if (declarations && declarations.length > 0) {
                                        const declaration = declarations[0];
                                        const declarationComments = (0, utils_1.getCommentsFromDeclaration)(declaration);
                                        if (declarationComments.length > 0) {
                                            log(`[IMPORT] Found comments on declaration: ${declarationComments.join("\n")}`);
                                        }
                                        if ((0, utils_1.isDeclarationDeprecated)(declaration)) {
                                            log(`[IMPORT] Import "${importName}" is deprecated in "${moduleName}"`);
                                            // Get the node of the named import within the original node
                                            const importNode = (_b = node
                                                .getChildren()
                                                .find(child => ts.isNamedImports(child))) === null || _b === void 0 ? void 0 : _b.getChildren().find(child => ts.isImportSpecifier(child) && child.name.text === importName);
                                            diagnostics.push(createDeprecatedDiagnostic(importNode, exportSymbol, node.getSourceFile(), [], log, declaration));
                                        }
                                    }
                                }
                                else {
                                    log(`[IMPORT] Could not find export "${importName}" in module "${moduleName}"`);
                                }
                            }
                        }
                        else {
                            log(`[IMPORT] Could not get symbol for module "${moduleName}"`);
                        }
                    }
                    else {
                        log(`[IMPORT] Could not get source file for "${resolvedFileName}"`);
                    }
                }
                else {
                    log(`[IMPORT] Could not resolve module "${moduleName}"`);
                }
            }
        }
        return diagnostics;
    }
};
// Resolve the symbol and follow alias chain to check for deprecations
const isSymbolDeprecatedRecursively = (symbol, checker, log) => {
    let currentSymbol = symbol;
    const visitedSymbols = new Set();
    while (currentSymbol) {
        // Avoid cycles
        if (visitedSymbols.has(currentSymbol))
            break;
        visitedSymbols.add(currentSymbol);
        // Check for deprecation in the current symbol
        if ((0, utils_1.isSymbolDeprecated)(currentSymbol)) {
            log(`Symbol ${currentSymbol.getName()} is deprecated.`);
            return true;
        }
        const isAlias = currentSymbol.flags & ts.SymbolFlags.Alias;
        if (!isAlias) {
            break;
        }
        log(`Symbol ${currentSymbol.getName()} is an alias.`);
        const immediateAliasedSymbol = checker.getImmediateAliasedSymbol(currentSymbol);
        if (immediateAliasedSymbol) {
            log(`Immediated alias from ${currentSymbol.getName()} to ${immediateAliasedSymbol.getName()}`);
        }
        // Resolve alias, if present, and continue the loop
        const aliasedSymbol = checker.getAliasedSymbol(currentSymbol);
        if (aliasedSymbol && aliasedSymbol !== currentSymbol) {
            log(`Following alias from ${currentSymbol.getName()} to ${aliasedSymbol.getName()}`);
            currentSymbol = aliasedSymbol;
        }
        else {
            break;
        }
    }
    return false;
};
// The factory function that TypeScript expects
const init = ({ typescript: ts }) => {
    return {
        create: info => {
            var _a;
            const proxy = Object.create(null);
            const oldLanguageService = info.languageService;
            const oldGetQuickInfoAtPosition = oldLanguageService.getQuickInfoAtPosition;
            const oldGetSemanticDiagnostics = oldLanguageService.getSemanticDiagnostics;
            const oldGetSuggestionDiagnostics = oldLanguageService.getSuggestionDiagnostics;
            // Create a proxy to wrap the language service
            for (let k in oldLanguageService) {
                const x = oldLanguageService[k];
                proxy[k] = (...args) => x.apply(oldLanguageService, args);
            }
            // Logging to TypeScript Server
            const log = (message) => {
                // Or use the logger if available
                // if (info.project && info.project.projectService && info.project.projectService.logger) {
                //     info.project.projectService.logger.info(`[DEPRECATION PLUGIN]: ${message}`);
                // } else {
                //     console.log(`[DEPRECATION PLUGIN]: ${message}`);
                // }
            };
            log("Plugin Initialized - hotloading");
            const checker = (_a = info.languageService.getProgram()) === null || _a === void 0 ? void 0 : _a.getTypeChecker();
            // Hook into the quick info to display tooltips
            // proxy.getQuickInfoAtPosition = (fileName: string, position: number) => {
            //     if (!isSupportedFileType(fileName)) return oldGetQuickInfoAtPosition(fileName, position);
            //     const quickInfo: ts.QuickInfo = oldGetQuickInfoAtPosition(fileName, position);
            //     const sourceFile = info.languageService.getProgram()?.getSourceFile(fileName);
            //     const node = findNodeAtPosition(sourceFile, position);
            //     if (node && checker) {
            //         const symbol = checker.getSymbolAtLocation(node);
            //         if (symbol && isSymbolDeprecatedRecursively(symbol, checker, log)) {
            //             log(`Deprecated symbol detected: ${symbol.getName()}`);
            //             // Create detailed deprecation message with source information
            //             const deprecationTag = createDeprecatedQuickInfoTag(symbol, checker);
            //             if (deprecationTag) {
            //                 quickInfo.tags = quickInfo.tags || [];
            //                 quickInfo.tags.push(deprecationTag);
            //             }
            //         }
            //     }
            //     return quickInfo;
            // };
            // Common logic for both suggestion and semantic diagnostics
            const checkDiagnostics = (fileName, priorDiagnostics, category, program = info.languageService.getProgram()) => {
                var _a;
                const sourceFile = (_a = info.languageService.getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(fileName);
                const diagnostics = [];
                if (sourceFile && checker && (0, utils_1.isSupportedFileType)(fileName)) {
                    log(`Checking ${fileName} for deprecated symbols`);
                    const visit = (node) => {
                        var _a;
                        // First we check for imports
                        const deprecatedAliasDiagnostics = isImportDeclarationDeprecated(node, checker, log, program);
                        if (deprecatedAliasDiagnostics === null || deprecatedAliasDiagnostics === void 0 ? void 0 : deprecatedAliasDiagnostics.length) {
                            for (const diagnostic of deprecatedAliasDiagnostics) {
                                diagnostic.category = category;
                                diagnostics.push(diagnostic);
                            }
                            return;
                        }
                        else if (!ts.isImportDeclaration(node) && !ts.isExportDeclaration(node)) {
                            // This will check for deprecated symbols in the source file by checkin the import declarations using the same logic as above
                            const symbol = checker.getSymbolAtLocation(node);
                            if (symbol) {
                                // Find the first declaration associated with the symbol
                                const declaration = (_a = symbol.getDeclarations()) === null || _a === void 0 ? void 0 : _a[0];
                                if (declaration && ts.isImportSpecifier(declaration)) {
                                    log(`Declaration found in import: ${declaration.getText()}`);
                                    let importDeclaration = declaration;
                                    // Get top-level import declaration
                                    while (importDeclaration.parent) {
                                        if (ts.isImportDeclaration(importDeclaration.parent)) {
                                            importDeclaration = importDeclaration.parent;
                                            break;
                                        }
                                        importDeclaration = importDeclaration.parent;
                                    }
                                    // Check if the import is deprecated
                                    if (ts.isImportDeclaration(importDeclaration)) {
                                        log(`Import parent: ${importDeclaration.getText()}`);
                                        const deprecatedAliasDiagnostics = isImportDeclarationDeprecated(importDeclaration, checker, log, program, symbol);
                                        if (deprecatedAliasDiagnostics === null || deprecatedAliasDiagnostics === void 0 ? void 0 : deprecatedAliasDiagnostics.length) {
                                            for (const diagnostic of deprecatedAliasDiagnostics) {
                                                diagnostic.start = node.getStart();
                                                diagnostic.length = node.getEnd() - node.getStart();
                                                diagnostics.push(diagnostic);
                                            }
                                        }
                                    }
                                }
                            }
                        }
                        ts.forEachChild(node, visit);
                    };
                    visit(sourceFile);
                }
                return priorDiagnostics.concat(diagnostics);
            };
            // Override suggestion diagnostics
            proxy.getSuggestionDiagnostics = (fileName) => {
                const priorDiagnostics = oldGetSuggestionDiagnostics(fileName);
                return checkDiagnostics(fileName, priorDiagnostics, ts.DiagnosticCategory.Suggestion);
            };
            // Override semantic diagnostics
            proxy.getSemanticDiagnostics = (fileName) => {
                const priorDiagnostics = oldGetSemanticDiagnostics(fileName);
                return checkDiagnostics(fileName, priorDiagnostics, ts.DiagnosticCategory.Warning);
            };
            return proxy;
        }
    };
};
module.exports = init;
//# sourceMappingURL=index.js.map